from django.test import TestCase
from wagtail_editorjs.registry.element.element import EditorJSElement
from wagtail_editorjs.registry.element.utils import (
    wrap_tag,
    _make_attr,
    add_attributes,
    EditorJSElementAttribute,
    EditorJSStyleAttribute,
)


class TestEditorJSElement(TestCase):

    def test_element(self):
        element = EditorJSElement(
            "div",
            "Hello, World!",
            attrs = {
                "class": "test-class",
                "id": "test-id",
            }
        )

        self.assertHTMLEqual(
            str(element),
            '<div class="test-class" id="test-id">Hello, World!</div>'
        )

    def test_element_attrs(self):
        element = EditorJSElement(
            "div",
            "Hello, World!",
        )

        element = add_attributes(element, **{
            "class_": "test-class",
            "data-test": "test-data",
        })

        self.assertHTMLEqual(
            str(element),
            '<div class="test-class" data-test="test-data">Hello, World!</div>'
        )

    def test_make_attr(self):

        attrs = _make_attr({
            "color": "red",
            "background-color": "blue",
        })

        self.assertIsInstance(attrs, EditorJSStyleAttribute)
        self.assertEqual(
            str(attrs),
            'color: red;background-color: blue'
        )

        attrs = _make_attr("test-class")
        self.assertIsInstance(attrs, EditorJSElementAttribute)
        self.assertEqual(
            str(attrs),
            'test-class'
        )

    def test_wrap_tag(self):
        tag = wrap_tag(
            "div",
            {
                "class": "test-class",
                "id": "test-id",
            },
            "Hello, World!"
        )
    
        self.assertHTMLEqual(
            tag,
            '<div class="test-class" id="test-id">Hello, World!</div>'
        )
    
    def test_wrap_tag_styles(self):
        tag = wrap_tag(
            "div",
            {
                "id": "test-id",
                "class": ["test-class", "test-class-2"],
                "style": {
                    "color": "red",
                    "background-color": "blue",
                }
            },
            "Hello, World!"
        )

        self.assertHTMLEqual(
            tag,
            '<div id="test-id" class="test-class test-class-2" style="color: red;background-color: blue">Hello, World!</div>'
        )

    



from .blocks import (
    CodeFeature,
    DelimiterFeature,
    HeaderFeature,
    HTMLFeature,
    WarningFeature,
    TableFeature,
    BlockQuoteFeature,
    WagtailBlockFeature,
    EditorJSFeatureStructBlock,
    ButtonFeature,
)
from .lists import (
    NestedListFeature,
    CheckListFeature,
)
from .documents import (
    AttachesFeature,
)
from .images import (
    ImageFeature,
    ImageRowFeature,
)
from .inlines import (
    TooltipFeature,
    LinkFeature,
    LinkAutoCompleteFeature,
    DocumentFeature,
)
from .tunes import (
    AlignmentBlockTune,
    TextVariantTune,
    ColorTune,
    BackgroundColorTune,
)


from typing import Any, Union
from django import forms

from ..element import EditorJSElement
from ..value import EditorJSBlock
from .base import BaseEditorJSFeature


class EditorJSJavascriptFeature(BaseEditorJSFeature):
    def __init__(self, tool_name: str, js: Union[str, list[str]] = None, css: Union[str, list[str]] = None, weight: int = 0, allowed_tags: list[str] = None, allowed_attributes: dict[str, list[str]] = None):
        # 1 for klass - unused for this type of feature.
        super().__init__(tool_name, 1, js, css, None, {}, weight=weight, allowed_tags=allowed_tags, allowed_attributes=allowed_attributes)

    def get_config(self, context: dict[str, Any] = None) -> dict:
        """
            Javascript only features do not get access to any configuration.
            They are not passed into the EditorJS tools.
        """
        return None


class EditorJSFeature(BaseEditorJSFeature):

    def validate(self, data: Any):
        """
            Perform basic validation for an EditorJS block feature.
        """
        if not data:
            return
        
        if "data" not in data:
            raise forms.ValidationError("Invalid data format")
    
    def render_block_data(self, block: EditorJSBlock, context = None) -> "EditorJSElement":
        return EditorJSElement(
            "p",
            block["data"].get("text")
        )
    
    @classmethod
    def get_test_data(cls):
        return [
            {
                "text": "Hello, world!"
            }
        ]
    get_test_data.__doc__ = BaseEditorJSFeature.get_test_data.__doc__
    

class EditorJSTune(BaseEditorJSFeature):
    """
        Works mostly like EditorJSFeature, but is used for tunes.
        Handles validation differently.
    """

    def tune_element(self, element: "EditorJSElement", tune_value: Any, context = None) -> "EditorJSElement":
        """
            Perform any action on the element based on the data provided by the tune.
        """
        return element
    
    @classmethod
    def get_test_data(cls):
        """
            Currently automatic tests for tunes are unsupported.
        """
        return None



from django import (
    template,
    forms,
)
from ..render import render_editorjs_html
from ..registry import EDITOR_JS_FEATURES, EditorJSValue
from ..forms import _get_feature_scripts

register = template.Library()

@register.simple_tag(takes_context=True, name="editorjs")
def render_editorjs(context, data: EditorJSValue):
    return render_editorjs_html(data.features, data, context=context)


init_scripts = {
    "css": [
        'wagtail_editorjs/css/frontend.css',
    ]
}


@register.simple_tag(takes_context=False, name="editorjs_static")
def editorjs_static(type_of_script="css", features: list[str] = None):
    if type_of_script not in ["css", "js"]:
        raise ValueError("type_of_script must be either 'css' or 'js'")

    if features is None:
        features = EDITOR_JS_FEATURES.keys()

    feature_mapping = EDITOR_JS_FEATURES.get_by_weight(
        features
    )

    frontend_static = init_scripts.get(type_of_script, [])
    for feature in feature_mapping.values():

        frontend_static.extend(
            _get_feature_scripts(
                feature,
                f"get_frontend_{type_of_script}",
                list_obj=frontend_static,
            )
        )

    kwargs = {}
    if type_of_script == "css":
        kwargs["css"] = {
            "all": frontend_static,
        }
    elif type_of_script == "js":
        kwargs["js"] = frontend_static

    return forms.Media(**kwargs).render()





from .attrs import (
    EditorJSElementAttribute,
    EditorJSStyleAttribute,
)
from .utils import (
    wrap_tag,
    add_attributes,
)
from .element import (
    EditorJSElement,
    EditorJSSoupElement,
    EditorJSWrapper,
    wrapper,
)


"""
wagtail_editorjs
© 2024 Nigel van Keulen, goodadvice.it

A Wagtail integration for the Editor.js block editor.
Includes support for custom blocks, inline tools, and more.
Everything is dynamic and customizable.
"""

import distutils.version as pv



__version__ = 'v1.6.6'
VERSION = pv.LooseVersion(__version__)



from django.urls import path, include
from wagtail import hooks


@hooks.register("register_admin_urls")
def register_admin_urls():
    urls = []

    # Make sure all features are properly registered.
    from ..registry import EDITOR_JS_FEATURES
    EDITOR_JS_FEATURES._look_for_features()

    for hook in hooks.get_hooks("register_editorjs_urls"):
        urls += hook()

    return [
        path(
            'wagtail-editorjs/',
            name='wagtail_editorjs',
            view=include(
                (urls, 'wagtail_editorjs'),
                namespace='wagtail_editorjs'
            ),
        ),
    ]




from typing import Any
from django import forms
from django.utils.translation import gettext_lazy as _
from django.utils.functional import cached_property
from django.templatetags.static import static
from django.utils.safestring import mark_safe

import bleach

from ..registry import (
    PageChooserURLsMixin,
    EditorJSFeature,
    EditorJSBlock,
    EditorJSElement,
    EditorJSSoupElement,
    wrap_tag,
)



class CodeFeature(EditorJSFeature):
    allowed_tags = ["code"]
    allowed_attributes = ["class"]
    klass="CodeTool"
    js=[
        "wagtail_editorjs/vendor/editorjs/tools/code.js",
    ]

    def validate(self, data: Any):
        super().validate(data)

        if 'code' not in data['data']:
            raise forms.ValidationError('Invalid code value')
    
    def render_block_data(self, block: EditorJSBlock, context = None) -> EditorJSElement:
        return EditorJSElement("code", block["data"]["code"], attrs={"class": "code"})
    
    @classmethod
    def get_test_data(cls):
        return [
            {
                "code": "print('Hello, World!')",
            }
        ]


class DelimiterFeature(EditorJSFeature):
    allowed_tags = ["hr"]
    allowed_attributes = ["class"]
    klass="Delimiter"
    js = [
        "wagtail_editorjs/vendor/editorjs/tools/delimiter.js",
    ]

    def render_block_data(self, block: EditorJSBlock, context = None) -> EditorJSElement:
        return EditorJSElement("hr", close_tag=False, attrs={"class": "delimiter"})
    
    @classmethod
    def get_test_data(cls):
        return [{}, {}, {}]


class HeaderFeature(EditorJSFeature):
    allowed_tags = ["h1", "h2", "h3", "h4", "h5", "h6"]
    allowed_attributes = ["class"]
    klass="Header"
    js=[
        "wagtail_editorjs/vendor/editorjs/tools/header.js",
    ]

    def validate(self, data: Any):
        super().validate(data)

        level = data["data"].get("level")
        if level > 6 or level < 1:
            raise forms.ValidationError("Invalid level value")
    
    def render_block_data(self, block: EditorJSBlock, context = None) -> EditorJSElement:
        return EditorJSElement(
            "h" + str(block["data"]["level"]),
            block["data"].get("text")
        )
    
    @classmethod
    def get_test_data(cls):
        return [
            {
                "level": heading,
                "text": f"Header {heading}",
            } for heading in range(1, 7)
        ]


class HTMLFeature(EditorJSFeature):
    allowed_tags = bleach.ALLOWED_TAGS
    allowed_attributes = bleach.ALLOWED_ATTRIBUTES
    klass="RawTool"
    js = [
        "wagtail_editorjs/vendor/editorjs/tools/raw.js",
    ]

    def validate(self, data: Any):
        super().validate(data)

        if "html" not in data["data"]:
            raise forms.ValidationError("Invalid html value")
    
    def render_block_data(self, block: EditorJSBlock, context = None) -> EditorJSElement:
        return EditorJSElement("div", block["data"]["html"], attrs={"class": "html"})
    
    @classmethod
    def get_test_data(cls):
        return [
            {
                "html": "<p>This is a HTML block.</p>",
            }
        ]

class WarningFeature(EditorJSFeature):
    allowed_tags = ["div", "h2", "p"]
    allowed_attributes = ["class"]
    klass="Warning"
    js = [
        "wagtail_editorjs/vendor/editorjs/tools/warning.js",
    ]

    def validate(self, data: Any):
        super().validate(data)

        if "title" not in data["data"]:
            raise forms.ValidationError("Invalid title value")
        
        if "message" not in data["data"]:
            raise forms.ValidationError("Invalid message value")
    
    def render_block_data(self, block: EditorJSBlock, context = None) -> EditorJSElement:
        return EditorJSElement(
            "div",
            attrs={
                "class": "warning",
            },
            content=[
                EditorJSElement(
                    "h2",
                    block["data"]["title"],
                ),
                EditorJSElement(
                    "p",
                    block["data"]["message"],
                ),
            ]
        )
    
    @classmethod
    def get_test_data(cls):
        return [
            {
                "title": "Warning",
                "message": "This is a warning message.",
            }
        ]


class TableFeature(EditorJSFeature):
    allowed_tags = ["table", "tr", "th", "td", "thead", "tbody", "tfoot"]
    allowed_attributes = ["class"]
    klass="Table"
    js = [
        "wagtail_editorjs/vendor/editorjs/tools/table.js",
    ]

    def validate(self, data: Any):
        super().validate(data)

        if "content" not in data["data"]:
            raise forms.ValidationError("Invalid content value")
        
        if "withHeadings" not in data["data"]:
            raise forms.ValidationError("Invalid withHeadings value")

    def render_block_data(self, block: EditorJSBlock, context = None) -> EditorJSElement:
        table = []

        if block["data"]["withHeadings"]:
            headings = block["data"]["content"][0]
            table.append(EditorJSElement(
                "thead",
                EditorJSElement(
                    "tr",
                    [
                        EditorJSElement("th", heading)
                        for heading in headings
                    ]
                )
            ))

            content = block["data"]["content"][1:]
        else:
            content = block["data"]["content"]

        tbody = EditorJSElement("tbody")
        for row in content:
            tbody.append(
                EditorJSElement(
                    "tr",
                    [
                        EditorJSElement("td", cell)
                        for cell in row
                    ]
                )
            )

        table.append(tbody)

        return EditorJSElement("table", table)

    @classmethod
    def get_test_data(cls):
        return [
            {
                "withHeadings": False,
                "content": [
                    ["1", "2", "3"],
                    ["4", "5", "6"],
                    ["7", "8", "9"],
                ],
            },
            {
                "withHeadings": True,
                "content": [
                    ["Heading 1", "Heading 2", "Heading 3"],
                    ["1", "2", "3"],
                    ["4", "5", "6"],
                    ["7", "8", "9"],
                ],
            }
        ]


class BlockQuoteFeature(EditorJSFeature):
    allowed_tags = ["blockquote", "footer"]
    allowed_attributes = ["class"]
    klass="Quote"
    js = [
        "wagtail_editorjs/vendor/editorjs/tools/quote.js",
    ]

    def validate(self, data: Any):
        super().validate(data)

        if "text" not in data["data"]:
            raise forms.ValidationError("Invalid text value")
        
        if "caption" not in data["data"]:
            raise forms.ValidationError("Invalid caption value")
        
    
    def render_block_data(self, block: EditorJSBlock, context = None) -> EditorJSElement:
        text = block["data"]["text"]
        caption = block["data"]["caption"]
        return EditorJSElement(
            "blockquote",
            [
                text,
                wrap_tag("footer", {}, caption),
            ],
            {
                "class": "blockquote",
            }
        )
    
    @classmethod
    def get_test_data(cls):
        return [
            {
                "text": "This is a quote.",
                "caption": "Anonymous",
            }
        ]

from wagtail import blocks


class EditorJSFeatureStructBlock(blocks.StructBlock):
    MUTABLE_META_ATTRIBUTES = blocks.StructBlock.MUTABLE_META_ATTRIBUTES + [
        "allowed_tags", "allowed_attributes"
    ]

    def __init__(self, *args, allowed_tags: list[str] = None, allowed_attributes: dict[str, list[str]] = None, **kwargs):
        super().__init__(*args, **kwargs)

        self.set_meta_options({
            "allowed_tags": allowed_tags,
            "allowed_attributes": allowed_attributes,
        })

        for _, block in self.child_blocks.items():
            if isinstance(block, (blocks.StreamBlock, blocks.ListBlock)):
                raise ValueError("StreamBlock and ListBlock are not allowed in StructBlock children for the EditorJSFeatureStructBlock")

def get_allowed_tags(instance: EditorJSFeatureStructBlock, block: blocks.Block, l = None) -> list[str]:

    if hasattr(block, "allowed_tags"):
        l = block.allowed_tags
    elif hasattr(block.meta, "allowed_tags"):
        l = block.meta.allowed_tags

    l = l or []

    if isinstance(block, blocks.StructBlock):
        for _, b in block.child_blocks.items():
            l += get_allowed_tags(instance, b, l)

    return list(set(list(l) + ["div"]))

def get_allowed_attributes(instance: EditorJSFeatureStructBlock, block: blocks.Block, d = None) -> dict[str, list[str]]:
    if hasattr(block, "allowed_attributes"):
        obj_dict = block.allowed_attributes
    elif hasattr(block.meta, "allowed_attributes"):
        obj_dict = block.meta.allowed_attributes
    else:
        obj_dict = {}

    if d is None:
        d = obj_dict
    else:
        for key, value in obj_dict.items():
            d[key] = set(list(d.get(key, [])) + list(value))
    
    if isinstance(d, list) and "class" not in d:
        d.append("class")
    elif isinstance(d, dict):
        v = d.get("div", [])
        if isinstance(v, list) and "class" not in v:
            v.append("class")
        elif isinstance(v, set) and "class" not in v:
            v.add("class")
        elif isinstance(v, str):
            v = set(v.split(" "))
            v.add("class")

        d["div"] = v

    if isinstance(block, blocks.StructBlock):
        for _, b in block.child_blocks.items():
            d = get_allowed_attributes(instance, b, d)

    return d


class WagtailBlockFeature(EditorJSFeature):
    def __init__(self, 
            tool_name: str,
            block: blocks.Block,
            klass: str = None,
            config: dict = None,
            weight: int = 0, # Weight is used to manage which features.js files are loaded first.
            allowed_tags: list[str] = None,
            allowed_attributes: dict[str, list[str]] = None,
            **kwargs
        ):

        if not isinstance(block, blocks.Block) and issubclass(block, blocks.Block):
            block = block()


        self.block = block

        self.block.set_name(
            self.block.__class__.__name__.lower()
        )

        super().__init__(
            tool_name, klass, None, None, None, config, weight, allowed_tags, allowed_attributes, **kwargs
        )

    def init_static(self, css, js):
        pass

    def init_attrs(self, allowed_tags, allowed_attributes):
        pass

    @cached_property
    def widget(self):
        return blocks.BlockWidget(self.block)

    def get_config(self, context: dict[str, Any] = None) -> dict:
        data = super().get_config(context)
        config = data.get("config", {})
        config["rendered"] = self.widget.render_with_errors(
            "__PREFIX__", self.block.get_default(),
        )
        data["config"] = config
        return data
    
    
    @cached_property
    def allowed_tags(self):
        return get_allowed_tags(self, self.block)

    @cached_property
    def allowed_attributes(self):
        return get_allowed_attributes(self, self.block)
    
    @property
    def js(self):
        return [
            *self.widget.media._js,
            mark_safe(f"<script src=\"{ static('wagtail_editorjs/js/tools/wagtail-block.js') }\" data-name=\"{ self.block.name }\" data-title=\"{ self.block.label }\"></script>"),
        ]
    
    def validate(self, data: Any):
        super().validate(data)

        if "block" not in data["data"]:
            raise forms.ValidationError("Invalid block value")
        
        try:
            data["data"]["block"] = self.block.get_prep_value(
                self.block.clean(self.block.to_python(data["data"]["block"]))
            )
        except blocks.StreamBlockValidationError as e:
            raise e
        except blocks.list_block.ListBlockValidationError as e:
            raise e
        except blocks.struct_block.StructBlockValidationError as e:
            raise e
        except Exception as e:
            raise e
        
    def value_for_form(self, value: dict) -> dict:
        value = super().value_for_form(value)
        value["data"]["block"] = self.block.get_form_state(
            self.block.to_python(value["data"]["block"])
        )
        return value
    
    def render_block_data(self, block: EditorJSBlock, context=None) -> EditorJSElement:
        value: blocks.StructValue = self.block.to_python(block["data"]["block"])
        return EditorJSSoupElement(f"<div class=\"{self.tool_name}\">{ self.block.render(value) }</div>")
    
    @classmethod
    def get_test_data(cls):
        return []
    
    @property
    def css(self):
        return self.widget.media._css
        
    @property
    def klass(self):
        return f"WagtailBlockTool_{ self.block.name }"

    @js.setter
    def js(self, data: Any): pass

    @css.setter
    def css(self, data: Any): pass

    @klass.setter
    def klass(self, data: Any): pass

    

from wagtail.models import Page
from wagtail.admin.widgets import AdminPageChooser


class ButtonFeature(PageChooserURLsMixin, EditorJSFeature):
    allowed_tags:       list[str]            = ["a"]
    allowed_attributes: dict[str, list[str]] = {
        "a": ["href", "class"]
    }
    chooser_class = AdminPageChooser
    model         = Page
    klass         = "PageButtonTool"
    js            = [
        "wagtail_editorjs/js/tools/wagtail-button-tool.js",
    ]

    @cached_property
    def widget(self):
        if self.chooser_class is None:
            return None
        
        return self.chooser_class()
    
    def validate(self, data: Any):
        super().validate(data)

        if "pageId" not in data["data"]:
            raise forms.ValidationError("Invalid id value")
        
        if "text" not in data["data"]:
            raise forms.ValidationError("Invalid text value")

    def get_config(self, context: dict[str, Any]):
        config = super().get_config() or {}
        config.setdefault("config", {})
        config["config"][
            "chooserId"
        ] = f"editorjs-{self.model._meta.model_name}-button-chooser-{context['widget']['attrs']['id']}"
        return config
    
    def render_block_data(self, block: EditorJSBlock, context=None) -> EditorJSElement:
        try:
            page = self.model.objects.get(id=block["data"]["pageId"])
        except self.model.DoesNotExist:
            return None

        request = None
        if context:
            request = context.get("request")

        anchor = EditorJSElement(
            "a",
            block["data"]["text"],
            {
                "href": page.get_url(request),
                "class": "editor-button",
            }
        )

        return EditorJSElement(
            "div", anchor, {"class": "editor-button-container"},
        )

    def render_template(self, context: dict[str, Any] = None):
        if not self.widget:
            return super().render_template(context)
        
        return self.widget.render_html(
            f"editorjs-{self.model._meta.model_name}-button-chooser-{context['widget']['attrs']['id']}",
            None,
            {
                "id": f"editorjs-{self.model._meta.model_name}-button-chooser-{context['widget']['attrs']['id']}"
            },
        )

    @classmethod
    def get_test_data(cls):
        pages = cls.model.objects.all()[:5]

        return [
            {
                "pageId": page.id,
                "text": page.title,
            }
            for page in pages
        ]





"""
Hook used to register features with the editorjs widget.
"""
REGISTER_HOOK_NAME = "register_wagtail_editorjs_features"

"""
Hook used to build the configuration for the editorjs widget.
Any additional configuration or overrides can be done here.
"""
BUILD_CONFIG_HOOK = "editorjs_widget_build_config"




from typing import Any
from django.template.loader import render_to_string
from django.utils.safestring import mark_safe
from bs4 import BeautifulSoup

from .base import BaseEditorJSTest
from wagtail_editorjs.render import render_editorjs_html
from wagtail_editorjs.registry import (
    EditorJSTune,
    EditorJSFeature,
    EditorJSElement,
    EDITOR_JS_FEATURES,
)


class TestEditorJSTune(EditorJSTune):
    allowed_attributes = {
        "*": ["data-testing-id"],
    }
    klass = 1

    def tune_element(self, element: EditorJSElement, tune_value: Any, context=None) -> EditorJSElement:
        element.attrs["data-testing-id"] = tune_value
        return element


# Create your tests here.
class TestEditorJSFeatures(BaseEditorJSTest):

    def setUp(self) -> None:
        super().setUp()
        self.tune = TestEditorJSTune(
            "test_tune_feature",
            None
        )
        EDITOR_JS_FEATURES.register(
            "test_tune_feature",
            self.tune,
        )

    def test_editorjs_features(self):

        html = []
        test_data = []
        for i, feature in enumerate(EDITOR_JS_FEATURES.features.values()):
            test_data_list = feature.get_test_data()
            if not isinstance(feature, (EditorJSFeature))\
                or not test_data_list:
                continue

            for j, data in enumerate(test_data_list):
                test_data_list[j] = {
                    "id": "test_id_{}_{}".format(i, j),
                    "type": feature.tool_name,
                    "data": data,
                    "tunes": {
                        "test_tune_feature": "test_id_{}_{}_tune".format(i, j)
                    }
                }
                
            test_data.extend(test_data_list)

            for data in test_data_list:
                if hasattr(feature, "render_block_data"):
                    tpl = feature.render_block_data(data)
                    tpl = self.tune.tune_element(tpl, data["tunes"]["test_tune_feature"])                    
                    html.append(tpl)

        rendered_1 = render_editorjs_html(
            EDITOR_JS_FEATURES.keys(),
            {"blocks": test_data},
            clean=False,
        )

        rendered_2 = render_to_string(
            "wagtail_editorjs/rich_text.html",
            {"html": mark_safe("\n".join([str(h) for h in html]))}
        )

        soup1 = BeautifulSoup(rendered_1, "html.parser")
        soup2 = BeautifulSoup(rendered_2, "html.parser")

        d1 = soup1.decode(False)
        d2 = soup2.decode(False)
        self.assertHTMLEqual(
            d1, d2,
            msg=(
                f"The rendered HTML for feature {feature} does not match the expected output.\n"
                "This might be due to a change in the rendering process.\n\n"
                "Expected: {expected}\n\n"
                "Got: {got}" % {
                    "expected": d1,
                    "got": d2,
                }
            )
        )

    def test_cleaned_editorjs_features(self):

        html = []
        test_data = []
        for i, feature in enumerate(EDITOR_JS_FEATURES.features.values()):
            test_data_list = feature.get_test_data()
            if not isinstance(feature, (EditorJSFeature))\
                or not test_data_list:
                continue

            for j, data in enumerate(test_data_list):
                test_data_list[j] = {
                    "id": "test_id_{}_{}".format(i, j),
                    "type": feature.tool_name,
                    "data": data,
                    "tunes": {
                        "test_tune_feature": "test_id_{}_{}_tune".format(i, j)
                    }
                }

            for data in test_data_list:
                if hasattr(feature, "render_block_data"):
                    tpl = feature.render_block_data(data)
                    tpl = self.tune.tune_element(tpl, data["tunes"]["test_tune_feature"])
                    html.append(tpl)

            test_data.extend(test_data_list)

        rendered = render_editorjs_html(
            EDITOR_JS_FEATURES.keys(),
            {"blocks": test_data},
            clean=True,
        )

        soup = BeautifulSoup(rendered, "html.parser")

        for i, data in enumerate(test_data):
            block = soup.find(attrs={"data-testing-id": data["tunes"]["test_tune_feature"]})

            if not block:
                self.fail(
                    f"Block with id {data['tunes']['test_tune_feature']} not found.\n"
                    "The tune might not have been properly applied. Check the test data.\n\n"
                    f"Test data: {data}\n\n"
                    f"Soup: {soup}"
                )

            feature = EDITOR_JS_FEATURES[data["type"]]
            element = feature.render_block_data(data)
            element = self.tune.tune_element(element, data["tunes"]["test_tune_feature"])

            soup_element = BeautifulSoup(str(element), "html.parser")

            self.assertHTMLEqual(
                str(block).replace("\n", "").strip(), str(soup_element).replace("\n", "").strip(),
                msg=(
                    f"Block with feature {feature} ({i}) does not match the expected output.\n"
                    "Something has gone wrong with the cleaning process.\n\n"
                    "Expected: {expected}\n\n"
                    "Got: {got}" % {
                        "feature": data['tunes']['test_tune_feature'],
                        "expected": str(soup_element).replace('\n', '').strip(),
                        "got": str(block).replace('\n', '').strip(),
                    }
                )
            )





from typing import Any
from django import forms
from django.utils.translation import gettext_lazy as _

from ..registry import (
    EditorJSTune,
    EditorJSElement,
    wrapper,
)


class AlignmentBlockTune(EditorJSTune):
    allowed_attributes = {
        "*": ["class"],
    }
    klass = "AlignmentBlockTune"
    js = [
        "wagtail_editorjs/vendor/editorjs/tools/text-alignment.js",
    ]

    def validate(self, data: Any):
        super().validate(data)
        alignment = data.get("alignment")
        if alignment not in ["left", "center", "right"]:
            raise forms.ValidationError("Invalid alignment value")
        
    def tune_element(self, element: EditorJSElement, tune_value: Any, context = None) -> EditorJSElement:
        element = super().tune_element(element, tune_value, context=context)
        element.add_attributes(class_=f"align-content-{tune_value['alignment'].strip()}")
        return element


class TextVariantTune(EditorJSTune):
    allowed_tags = ["div"]
    allowed_attributes = ["class"]
    klass = "TextVariantTune"
    js = [
        "wagtail_editorjs/vendor/editorjs/tools/text-variant-tune.js",
    ]

    def validate(self, data: Any):
        super().validate(data)
        if not data:
            return
        
        if data not in [
                "call-out",
                "citation",
                "details",
            ]:
            raise forms.ValidationError("Invalid text variant value")
        
    def tune_element(self, element: EditorJSElement, tune_value: Any, context = None) -> EditorJSElement:
        element = super().tune_element(element, tune_value, context=context)

        if not tune_value:
            return element
        
        if element.is_wrapped:
            element["class"] = f"text-variant-{tune_value}"

        return EditorJSElement(
            "div",
            element,
            attrs={"class": f"text-variant-{tune_value}"},
        )


class ColorTune(EditorJSTune):
    allowed_attributes = {
        "*": ["class", "style"],
    }
    js = [
        "wagtail_editorjs/js/tools/wagtail-color-tune.js",
    ]
    klass = "WagtailTextColorTune"

    def validate(self, data: Any):
        super().validate(data)
        if not data:
            return
        
        if not isinstance(data, dict):
            raise forms.ValidationError("Invalid color value")
        
        if "color" not in data:
            # Dont do anything
            return
        
        if not isinstance(data["color"], str):
            raise forms.ValidationError("Invalid color value")
        
        if not data["color"].startswith("#"):
            raise forms.ValidationError("Invalid color value")
        
    def tune_element(self, element: EditorJSElement, tune_value: Any, context = None) -> EditorJSElement:
        if "color" not in tune_value:
            return element
        
        return wrapper(
            element,
            attrs={
                "class": "wagtail-editorjs-color-tuned",
                "style": {
                    "--text-color": tune_value["color"],
                },
            },
        )

class BackgroundColorTune(ColorTune):
    klass = "WagtailBackgroundColorTune"

    def tune_element(self, element: EditorJSElement, tune_value: Any, context = None) -> EditorJSElement:

        if "color" not in tune_value:
            return element
                
        classname = [
            "wagtail-editorjs-color-tuned",
        ]

        attrs = {
            "class": classname,
            "style": {
                "--background-color": tune_value["color"],
            },
        }

        if tune_value.get("stretched", None):
            attrs["class"] = classname + ["bg-stretched"]
        
        return wrapper(
            element,
            attrs=attrs,
        )



from typing import Any, Union, TypeVar
from .attrs import EditorJSElementAttribute
from .utils import add_attributes, wrap_tag, _make_attr
import bs4


ElementType = TypeVar("ElementType", bound="EditorJSElement")


class EditorJSElement:
    """
        Base class for all elements.
    """
    
    def __init__(self, tag: str, content: Union[str, list[str]] = None, attrs: dict[str, EditorJSElementAttribute] = None, close_tag: bool = True):
        attrs = attrs or {}
        content = content or []

        if not isinstance(attrs, EditorJSElementAttribute):
            attrs = {
                key: _make_attr(value)
                for key, value in attrs.items()
            }

        if isinstance(content, str):
            content = [content]
                
        self.tag = tag
        self._content = content
        self.attrs = attrs
        self.close_tag = close_tag

    @property
    def is_wrapped(self):
        return False

    def __setitem__(self, key, value):
        if "key" in self.attrs:
            attrs = self.attrs[key]
            attrs.append(value)
            self.attrs[key] = attrs
        else:
            self.attrs[key] = _make_attr(value)

    def add_attributes(self, **attrs: Union[str, list[str], dict[str, Any]]):
        return add_attributes(self, **attrs)
    
    @property
    def content(self):
        if isinstance(self._content, list):
            return "\n".join([str(item) for item in self._content])
        return str(self._content)
    
    @content.setter
    def content(self, value):
        if isinstance(value, list):
            self._content = value
        else:
            self._content = [value]
    
    def append(self, element: "EditorJSElement"):
        self._content.append(element)

    def __str__(self):
        return wrap_tag(
            self.tag,
            attrs = self.attrs,
            content = self.content,
            close_tag = self.close_tag
        )
    

class EditorJSSoupElement(EditorJSElement):
    def __init__(self, raw_html: str):
        self.raw_html = raw_html
        self.soup = bs4.BeautifulSoup(raw_html, "html.parser")
        self.soupContent = self.soup.contents[0]

    @property
    def content(self):
        return str(self.soup)
    
    @content.setter
    def content(self, value):
        self.soup = bs4.BeautifulSoup(value, "html.parser")
        self.soupContent = self.soup.contents[0]

    @property
    def attrs(self):
        return self.soupContent.attrs            

    def __str__(self):
        return str(self.soup)
    
    def append(self, element: "EditorJSElement"):
        if isinstance(element, EditorJSSoupElement):
            self.soupContent.append(element.soup)
        elif isinstance(element, str):
            self.soupContent.append(element)
        elif isinstance(element, EditorJSElement):
            self.soupContent.append(str(element))
        else:
            raise TypeError(f"Invalid type {type(element)}")
        
    def add_attributes(self, **attrs: Union[str, list[str], dict[str, Any]]):
        for key, value in attrs.items():
            if key == "class_" or key == "class":
                classList = self.soupContent.get("class", [])
                if isinstance(value, str):
                    classList.append(value)
                elif isinstance(value, list):
                    classList.extend(value)
                self.soupContent["class"] = classList
            else:
                self.soupContent[key] = value


def wrapper(element: EditorJSElement, attrs: dict[str, EditorJSElementAttribute] = None, tag: str = "div"):

    if isinstance(element, EditorJSWrapper) or getattr(element, "is_wrapped", False):
        return add_attributes(element, **attrs)

    return EditorJSWrapper(element, attrs=attrs, tag=tag)


class EditorJSWrapper(EditorJSElement):
    @property
    def is_wrapped(self):
        return True
    
    @property
    def wrapped_element(self) -> Union[ElementType, list[ElementType]]:
        if len(self._content) > 1:
            return self._content

        return self._content[0]
    
    def __init__(self,
            content: Union[ElementType, list[ElementType]],
            attrs: dict[str, EditorJSElementAttribute] = None,
            close_tag: bool = True,
            tag: str = "div",
        ):
        
        if not isinstance(content, (list, tuple)):
            content = [content]

        for item in content:

            item_type = type(item)
            if item_type == EditorJSWrapper:
                raise ValueError(
                    "Cannot nest EditorJSWrapper elements\n"
                    "Please check if the element is already wrapped before re-wrapping.\n"
                )

            if not issubclass(item_type, EditorJSElement):
                raise ValueError(f"Expected EditorJSElement got {type(item)}")

        super().__init__(tag, content, attrs, close_tag)
